/*
 *  $Id: pvrusb2-firmware.c,v 1.3 2005/05/12 02:58:08 isely Exp $
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *  Copyright (C) 2004 Aurelien Alleaume <slts@free.fr>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/device.h>
#include <linux/firmware.h>
#include "pvrusb2-firmware.h"
#include "pvrusb2-hdw-internal.h"
#include "pvrusb2-hdw-low.h"
#include "pvrusb2-debug.h"

#define trace_firmware(...) pvr2_trace(PVR2_TRACE_FIRMWARE,__VA_ARGS__)

#define PVR2_FIRMWARE_ENDPOINT   0x02

/* size of a firmware chunk */
#define FIRMWARE_CHUNK_SIZE 0x2000

/* Various files we will load with firmware_entry */
#define FIRMWARE1_FILE  "pvrusb2.f1"
#define FIRMWARE2_FILE  "pvrusb2.f2"

/*
 * pvr2_upload_firmware1().
 *
 * Send the 8051 firmware to the device.  After the upload, arrange for
 * device to re-enumerate.
 *
 * NOTE : the pointer to the firmware data given by request_firmware()
 * is not suitable for an usb transaction.
 *
 */
int pvr2_upload_firmware1(struct pvr2_hdw *hdw)
{
    const struct firmware *fw_entry = 0;
    void  *fw_ptr;
    unsigned int pipe;
    int ret;
    u16 address;
    const char *firmware_file = FIRMWARE1_FILE;

    trace_firmware("pvr2_upload_firmware1");

    usb_settoggle(hdw->usb_dev, 0 & 0xf, !(0 & USB_DIR_IN), 0);
    usb_clear_halt(hdw->usb_dev, usb_sndbulkpipe(hdw->usb_dev, 0 & 0x7f));

    pipe = usb_sndctrlpipe(hdw->usb_dev, 0);
    ret = request_firmware(&fw_entry, firmware_file, &hdw->usb_dev->dev);

    if (ret) {
        pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "request_firmware failed for '%s'", firmware_file);
        return ret;
    }

    if (fw_entry->size != 0x2000){
        pvr2_trace(PVR2_TRACE_ERROR_LEGS,"wrong fx2 firmware size");
        release_firmware(fw_entry);
        return -ENOMEM;
    }

    fw_ptr = kmalloc(0x800, GFP_KERNEL);
    if (fw_ptr == NULL){
        release_firmware(fw_entry);
        return -ENOMEM;
    }

    /*
     * Write 0x01 to the CPUCS register at address e600 to stop the CPU
     * during firmware upload.
     */

    ((u8*) fw_ptr)[0] = 0x01;
    ret = usb_control_msg(hdw->usb_dev,pipe,0xa0,0x40,0xe600,
                          0,fw_ptr,1,HZ);

    /*
     * upload the firmware to address 0000-1fff in 2048 (=0x800) bytes chunk.
     */

    for(address = 0; address < fw_entry->size; address += 0x800) {
        memcpy(fw_ptr, fw_entry->data + address, 0x800);
        ret += usb_control_msg(hdw->usb_dev, pipe, 0xa0, 0x40, address,
                               0, fw_ptr, 0x800, HZ);
    }

    trace_firmware("Upload done, releasing device's CPU");

    /*
     * Release the CPU from reset state by writing 0x00 to CPUCS.
     */
    ((char*) fw_ptr)[0] = 0x00;
    ret += usb_control_msg(hdw->usb_dev, pipe, 0xa0, 0x40, 0xe600,
                           0, fw_ptr, 1, HZ);

    kfree(fw_ptr);
    release_firmware(fw_entry);

    trace_firmware("Upload done (%d bytes sent)",ret);

    /* We should have written 1 + 8192 + 1 bytes */
    return (ret == 8194) ? 0 : -EIO;
}


/*
 * pvr2_upload_firmware2()
 *
 * This uploads encoder firmware on endpoint 2.
 *
 */

int pvr2_upload_firmware2(struct pvr2_hdw *hdw)
{
    const struct firmware *fw_entry = 0;
    void  *fw_ptr;
    unsigned int pipe, fw_len, fw_done;
    int actual_length;
    int ret = 0;

    /* First prepare firmware loading */
    ret |= pvr2_write_u8(hdw, 0x0b, 1);
    ret |= pvr2_write_u16(hdw, 0xde00, 0);
    ret |= pvr2_write_u8(hdw, 0xeb, 1);
    ret |= pvr2_write_register(hdw, 0x0048, 0xffffffff); /* interrupt mask */
    ret |= pvr2_write_register(hdw, 0x9020, 0x88000000); /* gpio dir */
    ret |= pvr2_write_register(hdw, 0x900c, 0x08000000); /* gpio output state */
    ret |= pvr2_write_u8(hdw, 0xdd, 0);
    ret |= pvr2_write_register(hdw, 0xa064, 0x00000000); /* APU command */
    ret |= pvr2_write_register(hdw, 0x9020, 0x08040000); /* gpio dir */
    ret |= pvr2_write_register(hdw, 0x900c, 0x08000000); /* gpio output state */
    ret |= pvr2_write_register(hdw, 0x9058, 0xedffffff); /* VPU ctrl */
    ret |= pvr2_write_register(hdw, 0x9054, 0xfdffffff); /* reset hw blocks */
    ret |= pvr2_write_register(hdw, 0x07f8, 0x00080080); /* encoder SDRAM refresh */
    ret |= pvr2_write_register(hdw, 0x07fc, 0x1a000000); /* encoder SDRAM pre-charge */
    ret |= pvr2_write_register(hdw, 0x0700, 0x00000000); /* I2C clock */
    ret |= pvr2_write_register(hdw, 0xaa00, 0x00000000); /* unknown */
    ret |= pvr2_write_register(hdw, 0xaa04, 0x10780500); /* unknown */
    ret |= pvr2_write_register(hdw, 0xaa10, 0x00851400); /* unknown */
    ret |= pvr2_write_register(hdw, 0xaa18, 0x00008400); /* unknown */
    ret |= pvr2_write_u8(hdw, 0x52, 0);
    ret |= pvr2_write_u16(hdw, 0x0600, 0);

    if (ret) {
	pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "%s: firmware2 upload prep failed, ret=%d",hdw->name,ret);
        return ret;
    }

    /* Now send firmware */

    ret = request_firmware(&fw_entry, FIRMWARE2_FILE, &hdw->usb_dev->dev);

    if (ret) {
        pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "request_firmware failed for '%s'", FIRMWARE2_FILE);
        return ret;
    }

    fw_len = fw_entry->size;

    if (fw_len % FIRMWARE_CHUNK_SIZE) {
        pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "size of %s must be a multiple of 8192B", FIRMWARE2_FILE);
        release_firmware(fw_entry);
        return -1;
    }

    fw_ptr = kmalloc(FIRMWARE_CHUNK_SIZE, GFP_KERNEL);
    if (fw_ptr == NULL){
        release_firmware(fw_entry);
        pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "%s: failed to allocate memory for firmware2 upload",
                   hdw->name);
        return -ENOMEM;
    }

    pipe = usb_sndbulkpipe(hdw->usb_dev, PVR2_FIRMWARE_ENDPOINT);

    for (fw_done = 0 ; (fw_done < fw_len) && !ret ; fw_done += FIRMWARE_CHUNK_SIZE ) {
        int i;
        memcpy(fw_ptr, fw_entry->data + fw_done, FIRMWARE_CHUNK_SIZE);
        /* Usbsnoop log  shows that we must swap bytes... */
        for (i = 0; i < FIRMWARE_CHUNK_SIZE/4 ; i++)
            ((u32 *)fw_ptr)[i] = ___swab32(((u32 *)fw_ptr)[i]);

        ret |= usb_bulk_msg(hdw->usb_dev, pipe, fw_ptr, FIRMWARE_CHUNK_SIZE,
                            &actual_length, HZ);
        ret |= (actual_length != FIRMWARE_CHUNK_SIZE);
    }

    trace_firmware("upload of %s : %i / %i ", FIRMWARE2_FILE, fw_done,
                   fw_entry->size);

    kfree(fw_ptr);
    release_firmware(fw_entry);

    if (ret) {
	pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "%s: firmware2 upload transfer failure",hdw->name);
        return ret;
    }

    /* Finish upload */

    ret |= pvr2_write_register(hdw, 0x9054, 0xffffffff); /* reset hw blocks */
    ret |= pvr2_write_register(hdw, 0x9058, 0xe8ffffff); /* VPU ctrl */
    ret |= pvr2_write_u16(hdw, 0x0600, 0);

    if (ret) {
	pvr2_trace(PVR2_TRACE_ERROR_LEGS,
                   "%s: firmware2 upload post-proc failure",hdw->name);
    }
    return ret;
}



/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
