/*
 *  $Id: pvrusb2-hdw.c,v 1.17 2005/05/16 01:40:03 isely Exp $
 *
 *  Copyright (C) 2005 Mike Isely <isely@pobox.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#include <linux/errno.h>
#include <linux/string.h>
#include <linux/slab.h>
#include <asm/semaphore.h>
#include "pvrusb2-hdw.h"
#include "pvrusb2-hdw-low.h"
#include "pvrusb2-i2c.h"
#include "pvrusb2-tuner.h"
#include "pvrusb2-eeprom.h"
#include "pvrusb2-firmware.h"
#include "pvrusb2-hdw-internal.h"
#include "pvrusb2-msp34x5g.h"
#include "pvrusb2-saa7115.h"
#include "pvrusb2-encoder.h"
#include "pvrusb2-debug.h"

static int hfull = 1;
static int width = 720;

module_param(hfull, int, 1);
MODULE_PARM_DESC(hfull, "full height video ?");
module_param(width, int, 720);
MODULE_PARM_DESC(width, "video width : 720,480,352");

typedef int (*pvr2_ctl_set_func)(struct pvr2_hdw *,int ctl_id,int val);
typedef int (*pvr2_ctl_get_func)(struct pvr2_hdw *,int ctl_id);

struct pvr2_ctl_def {
    const char *name;
    pvr2_ctl_set_func set_func;
    pvr2_ctl_get_func get_func;
    int max_value;
    int min_value;
    int default_value;
    const char **value_defs_ptr;
    unsigned int value_defs_count;
};


static const char *control_values_srate[] = {
    [PVR2_CVAL_SRATE_48]   = "48KHz",
    [PVR2_CVAL_SRATE_44_1] = "44.1KHz",
};

static const char *control_values_audiobitrate[] = {
    [PVR2_CVAL_AUDIOBITRATE_384] = "384kb/s",
    [PVR2_CVAL_AUDIOBITRATE_320] = "320kb/s",
    [PVR2_CVAL_AUDIOBITRATE_256] = "256kb/s",
    [PVR2_CVAL_AUDIOBITRATE_224] = "224kb/s",
    [PVR2_CVAL_AUDIOBITRATE_192] = "192kb/s",
    [PVR2_CVAL_AUDIOBITRATE_160] = "160kb/s",
    [PVR2_CVAL_AUDIOBITRATE_128] = "128kb/s",
    [PVR2_CVAL_AUDIOBITRATE_112] = "112kb/s",
    [PVR2_CVAL_AUDIOBITRATE_96]  = "96kb/s",
    [PVR2_CVAL_AUDIOBITRATE_80]  = "80kb/s",
    [PVR2_CVAL_AUDIOBITRATE_64]  = "64kb/s",
    [PVR2_CVAL_AUDIOBITRATE_56]  = "56kb/s",
    [PVR2_CVAL_AUDIOBITRATE_48]  = "48kb/s",
    [PVR2_CVAL_AUDIOBITRATE_32]  = "32kb/s",
    [PVR2_CVAL_AUDIOBITRATE_VBR] = "VBR",
};

static const char *control_values_audioemphasis[] = {
    [PVR2_CVAL_AUDIOEMPHASIS_NONE]  = "None",
    [PVR2_CVAL_AUDIOEMPHASIS_50_15] = "50/15us",
    [PVR2_CVAL_AUDIOEMPHASIS_CCITT] = "CCITT J.17",
};

static const char *control_values_videostandard[] = {
    [PVR2_CVAL_VIDEOSTANDARD_NTSC_M]  = "NTSC-M",
    [PVR2_CVAL_VIDEOSTANDARD_SECAM_L] = "SECAM-L",
    [PVR2_CVAL_VIDEOSTANDARD_PAL_BG]  = "PAL-BG",
    [PVR2_CVAL_VIDEOSTANDARD_PAL_I]   = "PAL-I",
    [PVR2_CVAL_VIDEOSTANDARD_PAL_DK]  = "PAL-DK",
};

static const char *control_values_input[] = {
    [PVR2_CVAL_INPUT_TV]        = "television",  // xawtv needs this name
    [PVR2_CVAL_INPUT_RADIO]     = "radio",
    [PVR2_CVAL_INPUT_SVIDEO]    = "s-video",
    [PVR2_CVAL_INPUT_COMPOSITE] = "composite",
};

static const char *control_values_audiomode[] = {
    [PVR2_CVAL_AUDIOMODE_MONO]   = "Mono",
    [PVR2_CVAL_AUDIOMODE_STEREO] = "Stereo",
    [PVR2_CVAL_AUDIOMODE_SAP]    = "SAP",
    [PVR2_CVAL_AUDIOMODE_LANG1]  = "Lang1",
    [PVR2_CVAL_AUDIOMODE_LANG2]  = "Lang2",
};


#define VDEF(x) \
    .value_defs_ptr = x, \
    .value_defs_count = (sizeof(x)/sizeof(x[0]))

static int pvr2_ctl_set_chanprog_id(struct pvr2_hdw *hdw,int ctl_id,int value);
static int pvr2_ctl_get_chanprog_id(struct pvr2_hdw *hdw,int ctl_id);
static int pvr2_ctl_get_signal(struct pvr2_hdw *hdw,int ctl_id);
static int pvr2_ctl_get_streaming(struct pvr2_hdw *hdw,int ctl_id);

static struct pvr2_ctl_def control_defs[PVR2_CID_COUNT] =
{
    [PVR2_CID_BRIGHTNESS] = {
        .name = "Brightness",
        .min_value = 0,
        .max_value = 255,
        .default_value = 128,
    },
    [PVR2_CID_CONTRAST] = {
        .name = "Contrast",
        .min_value = 0,
        .max_value = 127,
        .default_value = 68,
    },
    [PVR2_CID_SATURATION] = {
        .name = "Saturation",
        .min_value = 0,
        .max_value = 127,
        .default_value = 64,
    },
    [PVR2_CID_HUE] = {
        .name = "Hue",
        .min_value = -128,
        .max_value = 127,
        .default_value = 0,
    },
    [PVR2_CID_VOLUME] = {
        .name = "Volume",
        .min_value = 0,
        .max_value = 65535,
        .default_value = 65535,
    },
    [PVR2_CID_BALANCE] = {
        .name = "Balance",
        .min_value = -32768,
        .max_value = 32767,
        .default_value = 0,
    },
    [PVR2_CID_BASS] = {
        .name = "Bass",
        .min_value = -32768,
        .max_value = 32767,
        .default_value = 0,
    },
    [PVR2_CID_TREBLE] = {
        .name = "Treble",
        .min_value = -32768,
        .max_value = 32767,
        .default_value = 0,
    },
    [PVR2_CID_MUTE] = {
        .name = "Mute",
        .min_value = 0,
        .max_value = 1,
        .default_value = 0,
    },
    [PVR2_CID_SRATE] = {
        .name = "Sample rate",
        .min_value = PVR2_CVAL_SRATE_MIN,
        .max_value = PVR2_CVAL_SRATE_MAX,
        .default_value = PVR2_CVAL_SRATE_48,
        VDEF(control_values_srate),
    },
    [PVR2_CID_AUDIOBITRATE] = {
        .name = "Audio Bitrate",
        .min_value = PVR2_CVAL_AUDIOBITRATE_MIN,
        .max_value = PVR2_CVAL_AUDIOBITRATE_MAX,
        .default_value = PVR2_CVAL_AUDIOBITRATE_224,
        VDEF(control_values_audiobitrate),
    },
    [PVR2_CID_AUDIOCRC] = {
        .name = "Audio CRC",
        .min_value = 0,
        .max_value = 1,
        .default_value = 1,
    },
    [PVR2_CID_AUDIOEMPHASIS] = {
        .name = "Audio Emphasis",
        .min_value = PVR2_CVAL_AUDIOEMPHASIS_MIN,
        .max_value = PVR2_CVAL_AUDIOEMPHASIS_MAX,
        .default_value = PVR2_CVAL_AUDIOEMPHASIS_NONE,
        VDEF(control_values_audioemphasis),
    },
    [PVR2_CID_VBR] = {
        .name = "Variable video bitrate",
        .min_value = 0,
        .max_value = 1,
        .default_value = 0,
    },
    [PVR2_CID_AVERAGEVIDEOBITRATE] = {
        .name = "Average video bitrate",
        .min_value = 1,
        .max_value = 20000000,
        .default_value = 6000000,
    },
    [PVR2_CID_PEAKVIDEOBITRATE] = {
        .name = "Peak video bitrate",
        .min_value = 1,
        .max_value = 20000000,
        .default_value = 6000000,
    },
    [PVR2_CID_VIDEOSTANDARD] = {
        .name = "Video Standard",
        .min_value = PVR2_CVAL_VIDEOSTANDARD_MIN,
        .max_value = PVR2_CVAL_VIDEOSTANDARD_MAX,
        .default_value = PVR2_CVAL_VIDEOSTANDARD_NTSC_M,
        VDEF(control_values_videostandard),
    },
    [PVR2_CID_INPUT] = {
        .name = "Video Source",
        .min_value = PVR2_CVAL_INPUT_MIN,
        .max_value = PVR2_CVAL_INPUT_MAX,
        .default_value = PVR2_CVAL_INPUT_TV,
        VDEF(control_values_input),
    },
    [PVR2_CID_AUDIOMODE] = {
        .name = "Audio Mode",
        .min_value = PVR2_CVAL_AUDIOMODE_MIN,
        .max_value = PVR2_CVAL_AUDIOMODE_MAX,
        .default_value = PVR2_CVAL_AUDIOMODE_STEREO,
        VDEF(control_values_audiomode),
    },
    [PVR2_CID_FREQUENCY] = {
        .name = "Tuner Frequency (Hz)",
        .min_value = 76000000L,
        .max_value = 801250000L,
        .default_value = 175250000L,
    },
    [PVR2_CID_HRES] = {
        .name = "Horizontal capture resolution",
        .min_value = 320,
        .max_value = 720,
        .default_value = 720,
    },
    [PVR2_CID_VRES] = {
        .name = "Vertical capture resolution",
        .min_value = 200,
        .max_value = 625,
        .default_value = 480,
    },
    [PVR2_CID_INTERLACE] = {
        .name = "Interlace mode",
        .min_value = 0,
        .max_value = 1,
        .default_value = 0,
    },
    [PVR2_CID_AUDIOLAYER] = {
        .name = "Audio Layer",
        .min_value = 0, // This is all a wild guess
        .max_value = 3,
        .default_value = 2, // This appears to be all that is supported
    },
    [PVR2_CID_CHANNEL] = {
        .name = "Channel",
        .min_value = 0,
        .max_value = FREQTABLE_SIZE,
        .default_value = 0,
    },
    [PVR2_CID_CHANPROG_ID] = {
        .name = "Channel Program ID",
        .min_value = 0,
        .max_value = FREQTABLE_SIZE,
        .default_value = 0,
        .set_func = pvr2_ctl_set_chanprog_id,
        .get_func = pvr2_ctl_get_chanprog_id,
    },
    [PVR2_CID_CHANPROG_FREQ] = {
        .name = "Channel Program Frequency",
        .min_value = 76000000L,
        .max_value = 801250000L,
        .default_value = 0L,
    },
    [PVR2_CID_SIGNAL_PRESENT] = {
        .name = "Signal Present",
        .min_value = 0,
        .max_value = 1,
        .get_func = pvr2_ctl_get_signal,
    },
    [PVR2_CID_STREAMING_ENABLED] = {
        .name = "Streaming Enabled",
        .min_value = 0,
        .max_value = 1,
        .get_func = pvr2_ctl_get_streaming,
    },
    [PVR2_CID_STREAMING_SUSPEND] = {
        .name = "Streaming suspend (debug tool)",
        .min_value = 0,
        .max_value = 1,
        .default_value = 0,
    },
};

#undef VDEF


const char *pvr2_config_get_name(enum pvr2_config cfg)
{
    switch (cfg) {
    case pvr2_config_empty: return "empty";
    case pvr2_config_mpeg: return "mpeg";
    case pvr2_config_vbi: return "vbi";
    case pvr2_config_radio: return "radio";
    }
    return "<unknown>";
}


struct usb_device *pvr2_hdw_get_dev(struct pvr2_hdw *hdw)
{
    return hdw->usb_dev;
}


unsigned long pvr2_hdw_get_sn(struct pvr2_hdw *hdw)
{
    return hdw->serial_number;
}


/*
 * pvr2_stop()
 *
 * Stop the stream. Call with lock held. After it, we can free all the buffer
 * without any danger, as long as we don't release the pvr->sem lock.
 */

static void pvr2_stop(struct pvr2_hdw *hdw)
{
    pvr2_trace(PVR2_TRACE_START_STOP,"/*--TRACE_STREAM--*/ disable");

    /* mask all interrupts */
    pvr2_write_register(hdw, 0x0048, 0xffffffff);
    /* stop capture on encoder side */
    pvr2_encoder_stop(hdw);
    /* stop capture */
    pvr2_write_u8(hdw, 0x37, 0);
    pvr2_write_u8(hdw, 0x44, 0);
    /* change some GPIO data */
    pvr2_write_register(hdw, 0x9020, 0x81040000);
    pvr2_write_register(hdw, 0x900c, 0x08000000);
}

/*
 * pvr2_start()
 *
 * Start the stream. Call with lock held.
 *
 */

PVR2_DECLARE_CMD(start_capture, 0, "\x36\x03\x00\x00\x00\x00\x00\x44");

static void pvr2_start(struct pvr2_hdw *hdw)
{
    pvr2_trace(PVR2_TRACE_START_STOP,"/*--TRACE_STREAM--*/ enable");

    /* unmask some interrupts */
    pvr2_write_register(hdw, 0x0048, 0xbfffffff);
    /* change some GPIO data */
    pvr2_write_register(hdw, 0x9020, 0x80040000);
    pvr2_write_register(hdw, 0x900c, 0x00000000);

    /* Start captutre (in MPEG) */
//    pvr2_do_command(pvr, &start_capture);
    pvr2_write_u8(hdw, 0x36, 0);
    pvr2_write_u8(hdw, 0x43, 0);
    pvr2_encoder_start(hdw);
}


static int pvr2_ctl_get_streaming(struct pvr2_hdw *hdw,int ctl_id)
{
    return  hdw->flag_streaming_enabled ? 1 : 0;
}


void pvr2_hdw_update_streaming(struct pvr2_hdw *hdw)
{
    int enable;
    enable = (hdw->flag_streaming_enabled &&
              (hdw->controls[PVR2_CID_STREAMING_SUSPEND].value == 0));
    if (!enable == !hdw->flag_streaming_on) return;
    if (!hdw->flag_ok) return;
    hdw->flag_streaming_on = enable;
    if (enable) {
	pvr2_start(hdw);
    } else {
	pvr2_stop(hdw);
    }
}


int pvr2_hdw_set_streaming_no_lock(struct pvr2_hdw *hdw,int enableFl)
{
    if (!hdw->flag_ok) return -EIO;
    if ((!enableFl) == (!hdw->flag_streaming_enabled)) return 0;
    hdw->flag_streaming_enabled = enableFl;
    pvr2_hdw_update_streaming(hdw);
    return 0;
}

int pvr2_hdw_set_streaming(struct pvr2_hdw *hdw,int enable_flag)
{
    int ret;
    if (!hdw->flag_ok) return -EIO;
    if ((!enable_flag) == (!hdw->flag_streaming_enabled)) return 0;
    down(&hdw->big_lock);
    ret = pvr2_hdw_set_streaming_no_lock(hdw,enable_flag);
    up(&hdw->big_lock);
    return ret;
}

int pvr2_hdw_set_stream_type_no_lock(struct pvr2_hdw *hdw,
                                     enum pvr2_config config)
{
    int streaming = hdw->flag_streaming_enabled;
    if (!hdw->flag_ok) return -EIO;
    pvr2_hdw_set_streaming_no_lock(hdw,0);
    hdw->config = config;
    pvr2_hdw_set_streaming_no_lock(hdw,streaming);
    return 0;
}

int pvr2_hdw_set_stream_type(struct pvr2_hdw *hdw,enum pvr2_config config)
{
    int ret;
    if (!hdw->flag_ok) return -EIO;
    down(&hdw->big_lock);
    ret = pvr2_hdw_set_stream_type_no_lock(hdw,config);
    up(&hdw->big_lock);
    return ret;
}

int pvr2_hdw_setup(struct pvr2_hdw *hdw)
{
    unsigned int idx;
    pvr2_trace(PVR2_TRACE_INIT,"pvr2_hdw_setup(hdw=0x%x) begin",
               (unsigned int)hdw);
    down(&hdw->big_lock); do {
        if (hdw->usb_intf->cur_altsetting->desc.bNumEndpoints == 0) {
            pvr2_upload_firmware1(hdw);
            break;
        }

        hdw->flag_ok = !0;
        pvr2_i2c_init(hdw);

        if (pvr2_upload_firmware2(hdw) && pvr2_upload_firmware2(hdw)){
            pvr2_trace(PVR2_TRACE_ERROR_LEGS,"device unstable!!");
            pvr2_hdw_render_useless(hdw);
            break;
        }

        for (idx = 0; idx < PVR2_CID_COUNT; idx++) {
            pvr2_hdw_set_ctl_value_internal(hdw,idx,
                                            control_defs[idx].default_value);
        }

        pvr2_reset_ctl_endpoints(hdw);

        pvr2_eeprom_analyze(hdw);

        pvr2_tuner_set_type(hdw);

        pvr2_eeprom_set_default_standard(hdw);

        pvr2_hdw_commit_ctl_internal(hdw);

        hdw->vid_stream = pvr2_stream_create();
        pvr2_trace(PVR2_TRACE_INIT,
                   "pvr2_hdw_setup: video stream is 0x%x",
                   (unsigned int)hdw->vid_stream);
        if (hdw->vid_stream) {
            pvr2_stream_setup(hdw->vid_stream,hdw->usb_dev,PVR2_VID_ENDPOINT);
        }

        hdw->flag_init_ok = !0;

    } while (0); up(&hdw->big_lock);
    pvr2_trace(PVR2_TRACE_INIT,
               "pvr2_hdw_setup(hdw=0x%x) done, ok=%d init_ok=%d",
               (unsigned int)hdw,hdw->flag_ok,hdw->flag_init_ok);
    return hdw->flag_init_ok;
}

/* Create and return a structure for interacting with the underlying
   hardware */
struct pvr2_hdw *pvr2_hdw_create(struct usb_interface *intf)
{
    struct pvr2_hdw *hdw;
    __u8 ifnum;

    hdw = kmalloc(sizeof(*hdw),GFP_KERNEL);
    pvr2_trace(PVR2_TRACE_INIT,"pvr2_hdw_create: hdw=0x%x",(unsigned int)hdw);
    if (!hdw) return hdw;
    memset(hdw,0,sizeof(*hdw));
    strlcpy(hdw->name,"pvrusb2",sizeof(hdw->name));
    hdw->tuner_type = -1;

    hdw->usb_intf = intf;
    hdw->usb_dev = interface_to_usbdev(intf);

    ifnum = hdw->usb_intf->cur_altsetting->desc.bInterfaceNumber;
    usb_set_interface(hdw->usb_dev,ifnum,0);

    init_MUTEX(&hdw->ctl_sem);
    init_MUTEX(&hdw->big_lock);

    return hdw;
}


/* Destroy hardware interaction structure */
void pvr2_hdw_destroy(struct pvr2_hdw *hdw)
{
    pvr2_trace(PVR2_TRACE_INIT,"pvr2_hdw_destroy: hdw=0x%x",(unsigned int)hdw);
    if (hdw->vid_stream) {
        pvr2_stream_destroy(hdw->vid_stream);
        hdw->vid_stream = 0;
    }
    if (hdw->flag_init_ok) {
        pvr2_i2c_done(hdw);
    }
    kfree(hdw);
}


int pvr2_hdw_init_ok(struct pvr2_hdw *hdw)
{
    return hdw->flag_init_ok;
}


int pvr2_hdw_dev_ok(struct pvr2_hdw *hdw)
{
    return (hdw && hdw->flag_ok);
}


/* Called when hardware has been unplugged */
void pvr2_hdw_disconnect(struct pvr2_hdw *hdw)
{
    pvr2_trace(PVR2_TRACE_INIT,"pvr2_hdw_disconnect(hdw=0x%x)",
               (unsigned int)hdw);
    down(&hdw->big_lock);
    pvr2_hdw_render_useless(hdw);
    hdw->flag_disconnected = !0;
    hdw->usb_dev = 0;
    hdw->usb_intf = 0;
    up(&hdw->big_lock);
}


static int pvr2_ctl_set_chanprog_id(struct pvr2_hdw *hdw,int ctl_id,int value)
{
    /* This is a special case; the value to store is to an array, and
       the element to select is determined by PVR_CID_CHANPROG_ID. */
    int id = hdw->controls[PVR2_CID_CHANPROG_ID].value;
    if ((id < 1) || (id > FREQTABLE_SIZE)) return 0;
    hdw->freqTable[id-1] = value;
    if (hdw->controls[PVR2_CID_CHANNEL].value == id) {
        /* If the current channel happens to be the slot we just set,
           then act like the current channel just got changed so we'll
           update that too. */
        hdw->controls[PVR2_CID_CHANNEL].dirty = !0;
    }
    return 0;
}


static int pvr2_ctl_get_chanprog_id(struct pvr2_hdw *hdw,int ctl_id)
{
    /* This is a special case; the value to return is from an array,
       and the element to select is determined by
       PVR_CID_CHANPROG_ID. */
    int id = hdw->controls[PVR2_CID_CHANPROG_ID].value;
    if ((id < 1) || (id > FREQTABLE_SIZE)) return 0;
    return hdw->freqTable[id-1];
}


/* Retrieve current value for a given control */
int pvr2_hdw_get_ctl_value(struct pvr2_hdw *hdw,unsigned int ctl_id)
{
    int ret = 0;

    if (ctl_id >= PVR2_CID_COUNT) return 0;
    down(&hdw->big_lock); do {
        if (control_defs[ctl_id].get_func) {
            ret = control_defs[ctl_id].get_func(hdw,ctl_id);
            break;
        }
        ret = hdw->controls[ctl_id].value;
    } while(0); up(&hdw->big_lock);
    return ret;
}


/* Return true if control is writable */
int pvr2_hdw_get_ctl_rw(struct pvr2_hdw *hdw,unsigned int ctl_id)
{
    if (ctl_id >= PVR2_CID_COUNT) return 0;
    if (control_defs[ctl_id].get_func && !control_defs[ctl_id].set_func) {
        return 0;
    }
    return !0;
}


/* Retrieve legal minimum value for a given control */
int pvr2_hdw_get_ctl_min_value(struct pvr2_hdw *hdw,unsigned int ctl_id)
{
    if (ctl_id >= PVR2_CID_COUNT) return 0;
    return control_defs[ctl_id].min_value;
}


/* Retrieve legal maximum value for a given control */
int pvr2_hdw_get_ctl_max_value(struct pvr2_hdw *hdw,unsigned int ctl_id)
{
    if (ctl_id >= PVR2_CID_COUNT) return 0;
    return control_defs[ctl_id].max_value;
}


/* Set current value for given control - this is just stored; the hardware
   isn't updated until the commit function is called. */
int pvr2_hdw_set_ctl_value_internal(struct pvr2_hdw *hdw,
                                    unsigned int ctl_id,int value)
{
    if (ctl_id >= PVR2_CID_COUNT) return -EINVAL;
    if (value < control_defs[ctl_id].min_value) return -EINVAL;
    if (value > control_defs[ctl_id].max_value) return -EINVAL;
    if (control_defs[ctl_id].set_func) {
        return control_defs[ctl_id].set_func(hdw,ctl_id,value);
    } else if (control_defs[ctl_id].get_func) {
        // If there's no "set" function yet there is still a "get"
        // function, then treat this as a read-only value.
        return -EINVAL;
    }
    hdw->controls[ctl_id].value = value;
    hdw->controls[ctl_id].dirty = !0;
    return 0;
}


/* Set current value for given control - this is just stored; the hardware
   isn't updated until the commit function is called. */
int pvr2_hdw_set_ctl_value(struct pvr2_hdw *hdw,unsigned int ctl_id,int value)
{
    int ret;
    down(&hdw->big_lock); do {
        ret = pvr2_hdw_set_ctl_value_internal(hdw,ctl_id,value);
    } while (0); up(&hdw->big_lock);
    return ret;
}


/* Retrieve string name for a given control value (returns a null pointer
   for any invalid combinations). */
const char *pvr2_hdw_get_ctl_value_name(struct pvr2_hdw *hdw,
                                        unsigned int ctl_id,
                                        int value)
{
    struct pvr2_ctl_def *cdef;
    if (ctl_id >= PVR2_CID_COUNT) return 0;
    cdef = control_defs + ctl_id;
    if (! cdef->value_defs_ptr) return 0;
    if (value >= cdef->value_defs_count) return 0;
    return cdef->value_defs_ptr[value];
}


/* Retrieve string name for given control */
const char *pvr2_hdw_get_ctl_name(struct pvr2_hdw *hdw,unsigned int ctl_id)
{
    if (ctl_id >= PVR2_CID_COUNT) return 0;
    return control_defs[ctl_id].name;
}


/* Commit all control changes made up to this point */
int pvr2_hdw_commit_ctl_internal(struct pvr2_hdw *hdw)
{
    unsigned int idx;
    int value;
    const char *ctl_name;
    const char *ctl_value;
    int commit_flag = 0;
    int streaming_enabled = hdw->flag_streaming_enabled;

    for (idx = 0; idx < PVR2_CID_COUNT; idx++) {
        if (!hdw->controls[idx].dirty) continue;
        if (!commit_flag) {
            commit_flag = !0;
        }
        value = hdw->controls[idx].value;
        ctl_name = control_defs[idx].name;
        if (control_defs[idx].value_defs_ptr) {
            if (value < control_defs[idx].value_defs_count) {
                ctl_value = control_defs[idx].value_defs_ptr[value];
            } else {
                ctl_value = "<out of range>";
            }
        } else {
            ctl_value = "<integer>";
        }
        pvr2_trace(PVR2_TRACE_CTL,"/*--TRACE_COMMIT--*/ \"%s\" <-- %d (%s)",
                   ctl_name,value,ctl_value);
    }

    if (!commit_flag) {
        /* Nothing has changed */
        return 0;
    }

    // There must be a priority ordering here.

    // Let's see if the channel changed and we have to update the frequency
    // because of it.  This setup means one can tune the receiver either by
    // just setting the channel (using the frequency table), or by directly
    // programming the frequency.  How do we resolve the obvious conflict
    // here?  The direct frequency takes priority; if directly set then we
    // commit that value and force the channel to zero which is interpreted
    // to mean "none".  If on the other hand we see that the channel has
    // been set and it's a legal value, then we copy that into the
    // frequency this.  The metaphor here is similar to when you tune your
    // digital radio:  You an either set a frequency directly or punch up a
    // pre-programmed station.  Either way a frequency is set, and if you
    // do use a preset, then the radio also shows you which preset it is -
    // until you override that by directly entering a new frequency.
    if (hdw->controls[PVR2_CID_FREQUENCY].dirty) {
        // Frequency has been directly set, so clear out the channel.
        hdw->controls[PVR2_CID_CHANNEL].value = 0;
    } else if (hdw->controls[PVR2_CID_CHANNEL].dirty) {
        int id = hdw->controls[PVR2_CID_CHANNEL].value;
        if ((id > 0) && (id <= FREQTABLE_SIZE)) {
            if (hdw->controls[PVR2_CID_FREQUENCY].value !=
                hdw->freqTable[id-1]) {
                hdw->controls[PVR2_CID_FREQUENCY].value =
                    hdw->freqTable[id-1];
                hdw->controls[PVR2_CID_FREQUENCY].dirty = !0;
            }
        }
    }

    // When video standard changes, reset the hres and vres values - but if
    // the user has pending changes there, then let the changes take
    // priority.
    if (hdw->controls[PVR2_CID_VIDEOSTANDARD].dirty) {
        if (!hdw->controls[PVR2_CID_VRES].dirty) {
            // Rewrite the vertical resolution to be appropriate to the video
            // standard that has been selected.
            int nvres = hdw->controls[PVR2_CID_VRES].value;
            switch (hdw->controls[PVR2_CID_VIDEOSTANDARD].value) {
            case PVR2_CVAL_VIDEOSTANDARD_NTSC_M:
                nvres = 480;
                break;
            default:
                nvres = 576;
            }
            if (nvres != hdw->controls[PVR2_CID_VRES].value) {
                hdw->controls[PVR2_CID_VRES].value = nvres;
                hdw->controls[PVR2_CID_VRES].dirty = !0;
            }
        }
        if (!hdw->controls[PVR2_CID_INTERLACE].dirty) {
            if (!hdw->controls[PVR2_CID_INTERLACE].value) {
                hdw->controls[PVR2_CID_INTERLACE].value = 0;
                hdw->controls[PVR2_CID_INTERLACE].dirty = !0;
            }
        }
    }

    if (hdw->controls[PVR2_CID_VIDEOSTANDARD].dirty) {
        hdw->controls[PVR2_CID_FREQUENCY].dirty = !0;
        pvr2_hdw_set_streaming_no_lock(hdw,0);
        pvr2_trace(PVR2_TRACE_CTL,
                   "/*---TRACE_CTL----*/ pvr2_decoder_set_norm");
        pvr2_decoder_set_norm(hdw);
    }

    if (hdw->controls[PVR2_CID_HRES].dirty ||
        hdw->controls[PVR2_CID_VRES].dirty) {
        pvr2_hdw_set_streaming_no_lock(hdw,0);
        pvr2_trace(PVR2_TRACE_CTL,
                   "/*---TRACE_CTL----*/ pvr2_decoder_set_size");
        pvr2_decoder_set_size(hdw);
    }

    if (hdw->controls[PVR2_CID_BRIGHTNESS].dirty ||
        hdw->controls[PVR2_CID_CONTRAST].dirty ||
        hdw->controls[PVR2_CID_SATURATION].dirty ||
        hdw->controls[PVR2_CID_HUE].dirty) {
        pvr2_trace(PVR2_TRACE_CTL,
                   "/*---TRACE_CTL----*/ pvr2_decoder_set_bcsh");
        pvr2_decoder_set_bcsh(hdw);
    }

    if (hdw->controls[PVR2_CID_SRATE].dirty) {
        pvr2_hdw_set_streaming_no_lock(hdw,0);
        pvr2_trace(PVR2_TRACE_CTL,
                   "/*---TRACE_CTL----*/ pvr2_decoder_set_audio");
        pvr2_decoder_set_audio(hdw);
    }

    if (hdw->controls[PVR2_CID_VIDEOSTANDARD].dirty ||
        hdw->controls[PVR2_CID_VRES].dirty ||
        hdw->controls[PVR2_CID_HRES].dirty ||
        hdw->controls[PVR2_CID_INTERLACE].dirty ||
        hdw->controls[PVR2_CID_VBR].dirty ||
        hdw->controls[PVR2_CID_AVERAGEVIDEOBITRATE].dirty ||
        hdw->controls[PVR2_CID_PEAKVIDEOBITRATE].dirty ||
        hdw->controls[PVR2_CID_AUDIOBITRATE].dirty ||
        hdw->controls[PVR2_CID_SRATE].dirty ||
        hdw->controls[PVR2_CID_AUDIOLAYER].dirty ||
        hdw->controls[PVR2_CID_AUDIOCRC].dirty ||
        hdw->controls[PVR2_CID_AUDIOEMPHASIS].dirty) {
        pvr2_hdw_set_streaming_no_lock(hdw,0);
        pvr2_trace(PVR2_TRACE_CTL,
                   "/*---TRACE_CTL----*/ pvr2_encoder_configure");
        pvr2_encoder_configure(hdw);
    }
    if (hdw->controls[PVR2_CID_VIDEOSTANDARD].dirty) {
        hdw->controls[PVR2_CID_FREQUENCY].dirty = !0;
        pvr2_hdw_set_streaming_no_lock(hdw,0);
        pvr2_trace(PVR2_TRACE_CTL,
                   "/*---TRACE_CTL----*/ pvr2_tuner_set_standard");
        pvr2_tuner_set_standard(hdw);
        pvr2_audio_set_standard(hdw);
    }

    if (hdw->controls[PVR2_CID_AUDIOMODE].dirty ||
        hdw->controls[PVR2_CID_INPUT].dirty) {
        pvr2_trace(PVR2_TRACE_CTL,
                   "/*---TRACE_CTL----*/ pvr2_audio_set_stereo");
        pvr2_audio_set_stereo(hdw);
    }

    if (hdw->controls[PVR2_CID_VOLUME].dirty ||
        hdw->controls[PVR2_CID_BALANCE].dirty ||
        hdw->controls[PVR2_CID_BASS].dirty ||
        hdw->controls[PVR2_CID_TREBLE].dirty ||
        hdw->controls[PVR2_CID_MUTE].dirty) {
        pvr2_trace(PVR2_TRACE_CTL,
                   "/*---TRACE_CTL----*/ pvr2_audio_setvolume");
        pvr2_audio_setvolume(hdw);
    }

    if (hdw->controls[PVR2_CID_INPUT].dirty) {
        pvr2_trace(PVR2_TRACE_CTL,
                   "/*---TRACE_CTL----*/ pvr2_decoder_set_input");
        pvr2_decoder_set_input(hdw);
    }

    if (hdw->controls[PVR2_CID_FREQUENCY].dirty) {
        pvr2_trace(PVR2_TRACE_CTL,
                   "/*---TRACE_CTL----*/ pvr2_tuner_set_freq");
        pvr2_tuner_set_freq(hdw);
    }

    for (idx = 0; idx < PVR2_CID_COUNT; idx++) hdw->controls[idx].dirty = 0;

    hdw->flag_streaming_enabled = streaming_enabled;
    pvr2_hdw_update_streaming(hdw);

    return 0;
}


int pvr2_hdw_commit_ctl(struct pvr2_hdw *hdw)
{
    down(&hdw->big_lock); do {
        pvr2_hdw_commit_ctl_internal(hdw);
    } while (0); up(&hdw->big_lock);
    return 0;
}


/* Find out how many controls there are.  Legal ids are numbered from 1
   through this value. */
unsigned int pvr2_hdw_get_ctl_count(struct pvr2_hdw *hdw)
{
    return PVR2_CID_COUNT;
}


/* Return bit mask indicating signal status */
unsigned int pvr2_hdw_get_signal_status_internal(struct pvr2_hdw *hdw)
{
    unsigned int msk = 0;
    switch (hdw->controls[PVR2_CID_INPUT].value) {
    case PVR2_CVAL_INPUT_TV:
    case PVR2_CVAL_INPUT_RADIO:
        if (pvr2_decoder_is_tuned(hdw)) {
            msk |= PVR2_SIGNAL_OK;
            if (pvr2_audio_update_status(hdw) == 0) {
                if (hdw->flag_stereo) msk |= PVR2_SIGNAL_STEREO;
                if (hdw->flag_bilingual) msk |= PVR2_SIGNAL_SAP;
            }
        }
        break;
    default:
        msk |= PVR2_SIGNAL_OK | PVR2_SIGNAL_STEREO;
    }
    return msk;
}


static int pvr2_ctl_get_signal(struct pvr2_hdw *hdw,int ctl_id)
{
    return (pvr2_hdw_get_signal_status_internal(hdw) & PVR2_SIGNAL_OK) ? 1 : 0;
}

/* Return bit mask indicating signal status */
unsigned int pvr2_hdw_get_signal_status(struct pvr2_hdw *hdw)
{
    unsigned int msk = 0;
    down (&hdw->big_lock); do {
        msk = pvr2_hdw_get_signal_status_internal(hdw);
    } while (0); up(&hdw->big_lock);
    return msk;
}


/* Get handle to video output stream */
struct pvr2_stream *pvr2_hdw_get_video_stream(struct pvr2_hdw *hp)
{
    return hp->vid_stream;
}


/*
  Stuff for Emacs to see, in order to encourage consistent editing style:
  *** Local Variables: ***
  *** mode: c ***
  *** fill-column: 75 ***
  *** c-basic-offset: 4 ***
  *** indent-tabs-mode: nil ***
  *** End: ***

  Similarly for vim to see:
  vim:expandtab:textwidth=75

*/
